//Main.c
//
//Tasks:
//	Receive DMX code at 250kbps
//	Decode DMX bytes
//	Convert to useful format
//	Transmit to motor board at 250kbps
//
//	The software detects the onset of the DMX BREAK signal and a state machine takes over reception of
//	the incoming DMX data stream.  This data is then converted into a format compatible with the 
//	Intellabeam motor board and the data is sent.

//Personality settings:
//1 - Preheat.  	Bulb minimum is 5% power, see Update_Dimmer().
//2 - Bulb save.	Bulb maximum is 90% power, see Update_Dimmer().
//3 - Invert pan
//4 - Invert tilt
//5 - 
//6 - 
//7 - 
//8 - 

//This code essentially working: 28/01/06
//DMX ok (see below), Comms ok to motor board, Conversion routines ok
//Changing DIP switches from Addr 00000001 or PERS 00000000 corrupts DMX data.
//Pull up resistor packs are on the wrong side of the diodes!
//Update:  RBPU cures the PERS dis switch problem, but the DMX address selection still does not work correctly.

//29/01/06
//Now fully working.  The routine that counts the bytes before asserting VALID_DATA did not read the
//RCREG register and therefore did not reset the USART interupt correctly.  PIR1bits.RCIF=0 does not
//reset it - dummy=RCREG; does.



#define DEBUGPIN PORTAbits.RA5	
//I/O allocations
//Pinouts
#define Data_Rec_LED PORTA,0
//RA0	LED 1
//RA1	LED 2
#define Data_Error_LED PORTA,1
//RA2	
//RA3	
//RA4	
//RA5	

// NB: ENABLE RBPU RESISTORS
//RB0	ZERO CROSS INPUT
//RB1	NC
//RB2	SWITCH BIT 0
//RB3	SWITCH BIT 1
//RB4	SWITCH BIT 2
//RB5	SWITCH BIT 3
//RB6	ICD
//RB7	ICD

//RC0	/SWITCH	1 SELECT
//RC1	/SWITCH	2 SELECT
//RC2	DIMMER CONTROL OUTPUT
//RC3	/SWITCH	3 SELECT
//RC4	/SWITCH	4 SELECT
//RC5	RESET MOTORS OUTPUT
//RC6	TX TO MOTOR BOARD
//RC7	RX FROM	DMX INPUT

//RD0	LCD
//RD1	LCD
//RD2	LCD
//RD3	LCD
//RD4	LCD
//RD5	LCD
//RD6	LCD
//RD7	LCD

//RE0	LCD RS
//RE1	LCD WR
//RE2	LCD E

#include	<p18f452.h>								// Register definitions
#include 	<stdlib.h>
#include 	<string.h>
#include 	<i2c.h>									// I2C library functions
#include 	<pwm.h>									// PWM library functions
#include 	<adc.h>									// ADC library functions
#include	<portb.h>								// PORTB library function
#include 	<timers.h>								// Timer library functions
#include 	<delays.h>

#include	"main.h"

rom char VERSION_NUMBER[]= "V1.0";		//Max 7char string

const rom unsigned char tbl_07c7[]={0x81, 0x71, 0x61, 0x51, 0x41, 0x31, 0x21, 0x11, 0x10, 0x20, 0x30, 0x40, 0x50, 0x60, 0x70, 0x80, 0x0b, 0x0a, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00};
const rom unsigned char tbl_07e7[]={0x80, 0x70, 0x60, 0x50, 0x40, 0x30, 0x20, 0x10, 0x11, 0x21, 0x31, 0x41, 0x51, 0x61, 0x71, 0x81, 0x0b, 0x0a, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00};
//revised table for s-curve dimming of quartz lamp
const rom unsigned char tbl_scurve[]={0x08, 0x13, 0x19, 0x20, 0x24, 0x2a, 0x2d, 0x32, 0x34, 0x38, 0x3b, 0x3e, 0x41, 0x43, 0x45, 0x47, 0x4a, 0x4b, 0x4e, 0x50, 0x51, 0x53, 0x55, 0x57, 0x58, 0x5a, 0x5b, 0x5d, 0x5e, 0x60,
	0x62, 0x63, 0x64, 0x66, 0x67, 0x68, 0x6a, 0x6b, 0x6c, 0x6e, 0x6f, 0x70, 0x71, 0x73, 0x74, 0x75, 0x76, 0x78, 0x79, 0x7a, 0x7c, 0x7d, 0x7e, 0x7f, 0x81, 0x82, 0x83, 0x84, 0x86, 0x87,
	0x88, 0x8a, 0x8b, 0x8d, 0x8e, 0x8f, 0x90, 0x92, 0x93, 0x95, 0x97, 0x98, 0x9a, 0x9b, 0x9d, 0x9e, 0xa0, 0xa1, 0xa3, 0xa6, 0xa7, 0xa9, 0xab, 0xae, 0xaf, 0xb2, 0xb4, 0xb7, 0xb9, 0xbd,
	0xc1, 0xc3, 0xc8, 0xcb, 0xd1, 0xd5, 0xdd, 0xe3, 0xee, 0xff};
	
//---------------------------------------------------------------------
// interrupt routines
//---------------------------------------------------------------------
#pragma code hiVector = 0x08
void Interrupt_Entry(void)
{
	_asm GOTO	isr_handler _endasm
}
#pragma code

#pragma interrupt isr_handler

void isr_handler(void)
{
	while (PIR1bits.RCIF) 
	{
		PIR1bits.RCIF=0;
		CheckUsart();
	}
}

void CheckUsart(void)
{
	PORTAbits.RA2=0;
	PORTAbits.RA3=1;	//Set data RX LED on
	
	//IH_Check_HW_Overrun
	if(RCSTAbits.OERR)
	{
		//IH_HW_Overrun
		RCSTAbits.CREN=0;
		RCSTAbits.CREN=1;		//Reset UART
		Flag.DMX_ERROR=1;		//Set flag
		PORTAbits.RA2=1;		//Set data error LED on
		PORTAbits.RA3=0;
	}
	//IH_Check_9th_Bit
	if(RCSTAbits.RX9D)
	{
		//IH_Check_SW_Overrun
		if(RCSTAbits.FERR)
		{
			RCSTAbits.CREN=0;
			RCSTAbits.CREN=1;		//Reset UART
			Flag.DMX_ERROR=1;		//Set flag
			PORTAbits.RA2=1;		//Set data error LED on
			PORTAbits.RA3=0;
		}
		//IH_Test_Flags
		if(Iflag.VALID_CHANNEL)
		{
			//IH_VALID_CHANNEL
			dmx_buffer[IH_Channels_Received]=RCREG;
			IH_Channels_Received++;
			if(IH_Channels_Received==CHANNELS)
			{
				Iflag.BYTE=0;
				//Iflag.VALID_CHANNEL=0;	//Get ready to start again...
				//Iflag.START_DETECTED=0;
				//Iflag.BREAK_DETECTED=0;
				PORTAbits.RA2=0;
				PORTAbits.RA3=1;	//Set data RX LED on
				Flag.DMX_DATA=1;
			}

			return;
		}
		if(Iflag.START_DETECTED)
		{
			//IH_WAIT_CHANNEL
			IH_Channel_Counter++;
			dummy=RCREG;	//Dummy read to clear RCIF
	
			if(IH_Channel_Counter!=DMX_Start_Addr) return; 	//Exit
			Iflag.VALID_CHANNEL=1;
			dmx_buffer[0]=RCREG;
			IH_Channels_Received++;
			return;
		}
		if(Iflag.BREAK_DETECTED)
		{
			//IH_WAIT_START
			IH_Channels_Received=0;
			IH_Channel_Counter=0;
			if(RCREG==0)
			{
				Iflag.START_DETECTED=1;
			}else{
				Iflag.BYTE=0;
				//Iflag.START_DETECTED=0;
				//Iflag.BREAK_DETECTED=0;
				//Iflag.VALID_CHANNEL=0;
			}
			return;			//Exit
		}
	}
	
	//IH_Wait_Break
	Iflag.BYTE=0;
	//Iflag.START_DETECTED=0;
	//Iflag.BREAK_DETECTED=0;
	//Iflag.VALID_CHANNEL=0;
	if((RCREG==0) && (RCSTAbits.RX9D==0))
	{
		Iflag.BREAK_DETECTED=1;
		Flag.DMX_DATA=0;
	}
	return;
	
}
//---------------------------------------------------------------------
// Main
//---------------------------------------------------------------------

void main(void)
{
	
	ADCON1 = 0x06;										//Port A and E are digital outputs
	TRISA  = 0;
	TRISB  = 0xff;
	TRISC  = 0x80;	
	TRISD  = 0;
	TRISE  = 0;	
	PORTC  = 0xff;
	//Setup timer 1 for 1:1 prescale, oscillator off, internal clock, stopped.
	//0xffff rollover at 5MHz = 200nS * 16777214 = 3.35 seconds
	//T1CON=0;
	//T1CONbits.T1CKPS1=1;	//Now set for 1:4 prescale
	//T1CONbits.TMR1ON=1;		//And switch timer on
	
	//Setup Timer0 
	//For 1sec, tick rate sould be approx 15uS (65536 * 15us) = 1sec
	//Clock rate is 100nS so prescale around 150 would be ok.
	//Nearest prescale is 128 giving 12.8uS (*65536=839mS).
	T0CON=0x06;
	T0CONbits.TMR0ON=1;		//Timer on
	
	SPBRG = 9;											// 250K baud @ 40MHz
	//SPBRG = 4;											// 250K baud @ 20MHz
	
	TXSTAbits.BRGH=1;									// Fast mode
	TXSTAbits.SYNC=0;
	TXSTAbits.TXEN=1;
	PIE1bits.RCIE=1;
	RCSTAbits.RX9=1;
	RCSTAbits.CREN=1;
	RCSTAbits.SPEN=1;
	
	//Setup timer2 for PWM output
	//FCLK=20MHz, FOSC/4 = 5MHz, Prescale = /4, Postscale NOT USED
	//TMR2_CLK = 1250KHz
	//8 bit, overflow = 1250KHz/256 = 4882.8Hz
	//PWM period = (PR2+1) * 4 * Tosc * (TMR2 prescale value)
	T2CON=0x05;	//Prescale 4, T2 on
	PR2=0xff;	//4.88KHz PWM frequency (approx 10KHz for 40MHz)
		
	//Setup PWM
	CCP1CON=0x0f;		//PWM mode on
	CCPR1L=0;			//Output at 0
	
	INTCON=0;	
	INTCONbits.PEIE=1;
	INTCONbits.GIE=1;
	
	//Enable RB pull-up resistors
	INTCON2bits.RBPU=0;	//Enable
	
	PORTCbits.RC5=1;
	//End reset motor board
	//Delay10KTCYx(50); //200nS * 500000 = 100mS
	Delay10KTCYx(100); //100nS * 1000000 = 100mS
	PORTCbits.RC5=0;
	
	
	while(1)
	{
		Read_Dip_Switches();
		dummy=TMR0L;	//Load buffered version of TMR0H for reading...
		if(TMR0H>0x7f) PORTAbits.RA5=0; else PORTAbits.RA5=1;

		//wait for an interrupt
		if(Flag.DMX_DATA)
		{
			diff=0;
			for(i=0; i<8; i++)
			{
				if(dmx_buffer[i]!=last_dmx[i]) diff=1;
				last_dmx[i]=dmx_buffer[i];
			}
			if(!diff)
			{
				Flag.DMX_DATA=0;				//Don't repeat call
				PORTAbits.RA0=1;
				//if(dmx_buffer[0]>0x7f)
				//{
				//	PORTAbits.RA1=1;
				//}else{
				//	PORTAbits.RA1=0;
				//}
				INTCONbits.GIE=0;			//Stop interrupts
				Convert_DMX();	//l0709
				Update_Motor_Board();	
				Update_Dimmer();
				INTCONbits.GIE=1;			//Enable interrupts
				PORTAbits.RA0=0;
			}
		}
	}
}
//---------------------------------------------------------------------
void Update_Dimmer(void)
{
	//value_dimmer = 0-255
	if(personality & 0x01)	//Preheat
	{
		if(value_dimmer<25) value_dimmer=25;
	}
	if(personality & 0x02)  //Bulb save
	{
		if(value_dimmer>230) value_dimmer=230;
	}
	
	CCPR1L = value_dimmer;
}
//---------------------------------------------------------------------
void Read_Dip_Switches(void)
{
//RB0	SWITCH BIT 0
//RB1	SWITCH BIT 1
//RB2	SWITCH BIT 2
//RB3	SWITCH BIT 3

//RC0	/SWITCH	1 SELECT
//RC1	/SWITCH	2 SELECT
//RC3	/SWITCH	3 SELECT
//RC4	/SWITCH	4 SELECT

	unsigned char temp;	
	temp=0;
	PORTCbits.RC3=1;	//personality lsb
	PORTCbits.RC4=1;	//personality msb
	
	PORTCbits.RC1=0;	//address msb
	PORTCbits.RC0=1;	//address lsb
	_asm
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		nop
	_endasm
	temp = (PORTB & 0x0f)<<4;
	PORTCbits.RC0=0;	//address lsb
	PORTCbits.RC1=1;	//address msb
	_asm
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		nop
	_endasm
	temp += (PORTB & 0x0f);
	PORTCbits.RC0=1;	//address lsb
	temp=255-temp;
	DMX_Start_Addr=(int)temp;
	
	PORTCbits.RC3=1;	//personality lsb
	PORTCbits.RC4=0;	//personality msb
	_asm
		nop
		nop
		nop
		nop
	_endasm
	personality = (PORTB & 0x0f)<<4;
	PORTCbits.RC3=0;	//personality lsb
	PORTCbits.RC4=1;	//personality msb
	_asm
		nop
		nop
		nop
		nop
	_endasm
	personality += (PORTB & 0x0f);
	PORTCbits.RC3=1;	//personality lsb
	personality=255-personality;
	//if(personality && 0x80) DMX_Start_Addr+=256;	//DMX 256-512
	
}
//---------------------------------------------------------------------
void Convert_DMX(void)
{
	int itemp;
	
	//Convert PAN
	if(personality & 0x04) dmx_buffer[0]=255-dmx_buffer[0];	//Invert
	int_value_pan = (int)dmx_buffer[0]*0x000c;				//PAN		0-3060
			
	//Convert TILT
	if(personality & 0x08) dmx_buffer[1]=255-dmx_buffer[1];	//Invert
	int_value_tilt= (int)dmx_buffer[1]*0x0004;				//TILT		0-1020
	
	//Convert COLOUR
	//int_ptr = ((int)dmx_buffer[2]*0x1c)>>8;				//COLOUR	0-0b from table
	//int_ptr &= 0x1f;
	//value_colour = tbl_07c7[int_ptr];
	itemp=((int)dmx_buffer[2]*0x000c)>>8;
	value_colour=0x0b-(unsigned char)itemp;
	
	//Convert GOBO
	//int_ptr = ((int)dmx_buffer[3]*0x1c)>>8;				//GOBO		0-0b from table
	//int_ptr &= 0x1f;
	//value_gobo = tbl_07e7[int_ptr];
	itemp=((int)dmx_buffer[3]*0x000c)>>8;
	value_gobo=0x0b-(unsigned char)itemp;
	
	//Convert SHUTTER
	itemp = ((int)dmx_buffer[4]*0x000a)>>8;		//SHUTTER	0=off, 1=on, 10-80 speeds
	value_shutter=(unsigned char)itemp;
	if(value_shutter!=0x09)
	{
		value_shutter<<=4;
	}else{
		value_shutter=0x0001;
	}
	
	//Convert DIMMER
	int_ptr=((int)dmx_buffer[5]*0x0064)>>8;				//DIMMER	0-255 s-curve
	value_dimmer=tbl_scurve[int_ptr];
	if(value_dimmer<0x10) value_shutter=0;
	
	//Convert IRIS
	itemp=((int)dmx_buffer[6]*0x0064)>>8;			//IRIS		0-100
	value_iris=(unsigned char)itemp;
	
	//Channel8
	//Software colour scrolling...
	if(dmx_buffer[7]>10)
	{
		int_scroll_colour+=dmx_buffer[7]>>2;
		value_colour=int_scroll_colour>>8;
		if(int_scroll_colour>0x0bff) int_scroll_colour=0;
	}
	//Set PAN/TILT speed according to size of change from last position
	//if(dmx_buffer[7] & 0xf8)
	//{
	//	pan_tilt_speed = ((int)dmx_buffer[7]*0x28)>>8;
	//	pan_tilt_speed += 0x3c;
	//}else{
		dmx=dmx_buffer[0];
		last_dmx_val=last_dmx_buffer[0];
		set_pan_tilt_speed();
		
		//if(pan_tilt_speed==0x63)
		//{
			dmx=dmx_buffer[1];
			last_dmx_val=last_dmx_buffer[1];
			set_pan_tilt_speed();
		//}
		last_dmx_buffer[0]=dmx_buffer[0];
		last_dmx_buffer[1]=dmx_buffer[1];
	//}
}
//---------------------------------------------------------------------
void set_pan_tilt_speed(void)
{
	//pan_tilt_speed:  92(0x5c)=slow, 99(0x63)=fast.
	if(last_dmx_val != dmx)
	{
		if(dmx>last_dmx_val)
		{
			if((dmx-last_dmx_val)>0x5)
			{
				pan_tilt_speed=0x63;
			}else{
				pan_tilt_speed=0x5c;
			}
		}else{
			if((last_dmx_val-dmx)>0x5)
			{
				pan_tilt_speed=0x63;
			}else{
				pan_tilt_speed=0x5c;
			}
		}
		last_dmx_val=dmx;
	}
}		
//---------------------------------------------------------------------
void Update_Motor_Board(void)
{
	Convert_Motor_Values();
	send_tx(0x10);
	Delay1KTCYx(MOTOR_DELAY);				//Delay for motor board
	for(i=0; i<14; i++)
	{
		send_tx(motor_buffer[i]);
	}
}
//---------------------------------------------------------------------
void Convert_Motor_Values(void)
{
	
	ltemp=int_value_pan & 0x0f00;
	ltemp>>=8;
	ltemp+=0x20;
	motor_buffer[0]=(unsigned char) ltemp;
	ltemp=int_value_pan & 0xf000;
	ltemp>>=12;
	ltemp+=0x20;
	motor_buffer[1]=(unsigned char) ltemp;
	ltemp=((int_value_pan & 0x000f)) + 0x20;
	motor_buffer[2]=(unsigned char) ltemp;
	ltemp=((int_value_pan & 0x00f0)>>4) + 0x20;
	motor_buffer[3]=(unsigned char) ltemp;
	
	ltemp=((int_value_tilt & 0x0f00)>>8) + 0x20;
	motor_buffer[4]=(unsigned char) ltemp;
	ltemp=((int_value_tilt & 0xf000)>>12) + 0x20;
	motor_buffer[5]=(unsigned char) ltemp;
	ltemp=((int_value_tilt & 0x000f)) + 0x20;
	motor_buffer[6]=(unsigned char) ltemp;
	ltemp=((int_value_tilt & 0x00f0)>>4) + 0x20;
	motor_buffer[7]=(unsigned char) ltemp;
	
	motor_buffer[8]=pan_tilt_speed+0x20;
	motor_buffer[9]=value_gobo+0x20;
	motor_buffer[10]=value_colour+0x20;
	motor_buffer[11]=value_iris+0x20;
	motor_buffer[12]=value_shutter+0x20;
	//if(Timer10ms>=50)
	//{
	//	motor_buffer[13]=0x21;	
	//}else{
		motor_buffer[13]=0x20;
	//}
}
//---------------------------------------------------------------------
void send_tx(unsigned char c)
{
	//while(PIR1bits.TXIF)			//Wait for last send to finish
	while(!TXSTAbits.TRMT)
	{
		PIR1bits.TXIF=0;
	}
	TXSTAbits.TX9=1;
	TXSTAbits.TX9D=1;
	TXREG=c;
	Delay1KTCYx(MOTOR_DELAY);				//Delay for motor board
	//FCLK=20MHz, Fcyc=5MHz, Tcyc=200nS
	//1KTCYx(4) = 4000 Tcyc = 800uS
}
//---------------------------------------------------------------------
